from __future__ import annotations
from dataclasses import is_dataclass, fields as dataclass_fields
from typing import Sequence
import typing as ty
from enum import Enum
import builtins
from copy import copy
import json
import re
import inspect
from importlib import import_module
from inspect import isclass, isfunction
from pathlib import PurePath, Path
import logging
import cloudpickle as cp
import attrs
from pydra.engine.core import Workflow, LazyField, TaskBase
from pydra.engine.task import FunctionTask
from arcana.core.exceptions import ArcanaUsageError
from .packaging import pkg_versions, package_from_module
from .misc import add_exc_note


logger = logging.getLogger("arcana")


ARCANA_PIP = "git+ssh://git@github.com/australian-imaging-service/arcana.git"

HASH_CHUNK_SIZE = 2**20  # 1MB in calc. checksums to avoid mem. issues


@attrs.define
class _FallbackContext:
    """Used to specify that class resolution is permitted to fail within this context
    and return just a string (i.e. in build environments where the required modules
    aren't installed)
    """

    permit: bool = False

    def __enter__(self):
        self.permit = True

    def __exit__(self, exception_type, exception_value, traceback):
        self.permit = False


@attrs.define
class ClassResolver:
    """
    Parameters
    ----------
    base_class : type
        the target class to resolve the string representation to
    prefixes : Sequence[str]
        List of allowable module prefixes to try to append if the fully
        resolved path fails, e.g. ['pydra.tasks'] would allow
        'fsl.preprocess.first.First' to resolve to
        pydra.tasks.fsl.preprocess.first.First
    """

    base_class: type = None
    allow_none: bool = False
    alternative_types: list[type] = attrs.field(factory=list)

    def __call__(self, class_str: str) -> type:
        """
        Resolves a class from a location string in the format "<module-name>:<class-name>"

        Parameters
        ----------
        class_str : str
            Module path and name of class joined by ':', e.g. main_pkg.sub_pkg:MyClass

        Returns
        -------
        type:
            The resolved class
        """
        if class_str is None and self.allow_none:
            return None
        klass = self.fromstr(class_str, subpkg=self._get_subpkg(self.base_class))
        self._check_type(klass)
        return klass

    @classmethod
    def _get_subpkg(cls, klass):
        try:
            return klass.SUBPACKAGE
        except AttributeError:
            return None

    @classmethod
    def fromstr(cls, class_str, subpkg=None):
        """Resolves a class/function from a string containing its module an its name
        separated by a ':'

        Parameters
        ----------
        class_str : str
            the string representation to resolve to a class or function
        subpkg : str, optional
            the sub-package that the class should belong to within the extension

        Returns
        -------
        type or callable
            the resolved class or function

        Raises
        ------
        ValueError
            raised if the string doesn't contain a ':'
        ArcanaUsageError
            raised if the class wasn't found in the sub-package
        ArcanaUsageError
            raised if a sub-package couldn't be found
        """
        if not isinstance(class_str, str):
            return class_str  # Assume that it is already resolved
        if class_str.startswith("<") and class_str.endswith(">"):
            class_str = class_str[1:-1]
        try:
            module_path, class_name = class_str.split(":")
        except ValueError:
            try:
                return getattr(builtins, class_str)
            except AttributeError:
                raise ValueError(
                    f"Class location '{class_str}' should contain a ':' unless it is in the "
                    "builtins module"
                ) from None

        if "." in module_path:
            # Interpret as an absolute path not a relative path from an extension
            module_path = module_path.rstrip(
                "."
            )  # trailing '.' signifies top-level pkg
            subpkg = None
        module = None

        if subpkg:
            full_mod_path = ".".join(("arcana", module_path, subpkg))
        else:
            full_mod_path = module_path
        try:
            module = import_module(full_mod_path)
        except ModuleNotFoundError:
            if cls.FALLBACK_TO_STR.permit:
                return class_str
            else:
                raise ArcanaUsageError(
                    f"Did not find module {full_mod_path}' when resolving {class_str} "
                    f"with subpkg={subpkg}"
                )
        try:
            klass = getattr(module, class_name)
        except AttributeError:
            raise ArcanaUsageError(
                f"Did not find '{class_str}' class/function in module '{module.__name__}'"
            )
        return klass

    @classmethod
    def tostr(cls, klass, strip_prefix: bool = True):
        """Records the location of a class so it can be loaded later using
        `ClassResolver`, in the format <module-name>:<class-name>

        Parameters
        ----------
        klass : Any
            the class/function to serialise to a string
        strip_prefix : bool
            whether to strip the SUBPACKAGE prefix from the module path when writing
            to file
        """
        if isinstance(klass, str):
            return klass
        if not (isclass(klass) or isfunction(klass)):
            klass = type(klass)  # Get the class rather than the object
        module_name = klass.__module__
        if module_name == "builtins":
            return klass.__name__
        if strip_prefix and cls._get_subpkg(klass):
            subpkg = cls._get_subpkg(klass)
            if match := re.match(r"arcana\.(\w+)\." + subpkg, module_name):
                module_name = match.group(
                    1
                )  # just use the name of the extension module
            elif "." not in module_name:
                module_name += "."  # To distinguish it from extension module name
        return module_name + ":" + klass.__name__

    def _check_type(self, klass):
        if self.FALLBACK_TO_STR.permit and isinstance(klass, str):
            return
        if self.base_class:
            if isfunction(klass):
                if ty.Callable in self.alternative_types:
                    return  # ok
                else:
                    raise ValueError(
                        f"Found callable {klass}, but Callable isn't in alternative_types"
                    )
            if klass in self.alternative_types:
                return  # ok
            if not isclass(klass) or not issubclass(klass, self.base_class):
                raise ValueError(
                    f"Found {klass}, which is not a subclass of {self.base_class}"
                )

    FALLBACK_TO_STR = _FallbackContext()


def asdict(obj, omit: ty.Iterable[str] = (), required_modules: set = None):
    """Serialises an object of a class defined with attrs to a dictionary

    Parameters
    ----------
    obj
        The Arcana object to asdict. Must be defined using the attrs
        decorator
    omit: Iterable[str]
        the names of attributes to omit from the dictionary
    required_modules: set
        modules required to reload the serialised object into memory"""

    def filter(atr, value):
        return atr.init and atr.metadata.get("asdict", True)

    if required_modules is None:
        required_modules = set()
        include_versions = True  # Assume top-level dictionary so need to include
    else:
        include_versions = False

    def serialise_class(klass):
        required_modules.add(klass.__module__)
        return "<" + ClassResolver.tostr(klass, strip_prefix=False) + ">"

    def value_asdict(value):
        if isclass(value):
            value = serialise_class(value)
        elif hasattr(value, "asdict"):
            value = value.asdict(required_modules=required_modules)
        elif attrs.has(value):  # is class with attrs
            value_class = serialise_class(type(value))
            value = attrs.asdict(
                value,
                recurse=False,
                filter=filter,
                value_serializer=lambda i, f, v: value_asdict(v),
            )
            value["class"] = value_class
        elif isinstance(value, Enum):
            value = serialise_class(type(value)) + "[" + str(value) + "]"
        elif isinstance(value, PurePath):
            value = "file://" + str(value.resolve())
        elif isinstance(value, (tuple, list, set, frozenset)):
            value = [value_asdict(x) for x in value]
        elif isinstance(value, dict):
            value = {value_asdict(k): value_asdict(v) for k, v in value.items()}
        elif is_dataclass(value):
            value = [
                value_asdict(getattr(value, f.name)) for f in dataclass_fields(value)
            ]
        return value

    dct = attrs.asdict(
        obj,
        recurse=False,
        filter=lambda a, v: filter(a, v) and a.name not in omit,
        value_serializer=lambda i, f, v: value_asdict(v),
    )

    dct["class"] = serialise_class(type(obj))
    if include_versions:
        dct["pkg_versions"] = pkg_versions(required_modules)

    return dct


def fromdict(dct: dict, **kwargs):
    """Unserialise an object from a dict created by the `asdict` method

    Parameters
    ----------
    dct : dict
        A dictionary containing a serialsed Arcana object such as a data store
        or dataset definition
    omit: Iterable[str]
        key names to ignore when unserialising
    **kwargs : dict[str, Any]
        Additional initialisation arguments for the object when it is reinitialised.
        Overrides those stored"""
    # try:
    #     arcana_version = dct["pkg_versions"]["arcana"]
    # except (TypeError, KeyError):
    #     pass
    #     else:
    #         if packaging.version.parse(arcana_version) < packaging.version.parse(MIN_SERIAL_VERSION):
    #             raise ArcanaVersionError(
    #                 f"Serialised version ('{arcana_version}' is too old to be "
    #                 f"read by this version of arcana ('{__version__}'), the minimum "
    #                 f"version is {MIN_SERIAL_VERSION}")

    def field_filter(klass, field_name):
        if attrs.has(klass):
            return field_name in (f.name for f in attrs.fields(klass))
        else:
            return field_name != "class"

    def fromdict(value):
        if isinstance(value, dict):
            if "class" in value:
                klass = ClassResolver()(value["class"])
                if hasattr(klass, "fromdict"):
                    return klass.fromdict(value)
            value = {fromdict(k): fromdict(v) for k, v in value.items()}
            if "class" in value:
                value = klass(
                    **{k: v for k, v in value.items() if field_filter(klass, k)}
                )
        elif isinstance(value, str):
            if match := re.match(r"<(.*)>$", value):  # Class location
                value = ClassResolver()(match.group(1))
            elif match := re.match(r"<(.*)>\[(.*)\]$", value):  # Enum
                value = ClassResolver()(match.group(1))[match.group(2)]
            elif match := re.match(r"file://(.*)", value):
                value = Path(match.group(1))
        elif isinstance(value, Sequence):
            value = [fromdict(x) for x in value]
        return value

    klass = ClassResolver()(dct["class"])

    kwargs.update(
        {
            k: fromdict(v)
            for k, v in dct.items()
            if field_filter(klass, k) and k not in kwargs
        }
    )

    return klass(**kwargs)


extract_import_re = re.compile(r"\s*(?:from|import)\s+([\w\.]+)")

NOTHING_STR = "__PIPELINE_INPUT__"


def pydra_asdict(
    obj: TaskBase, required_modules: ty.Set[str], workflow: Workflow = None
) -> dict:
    """Converts a Pydra Task/Workflow into a dictionary that can be serialised

    Parameters
    ----------
    obj : pydra.engine.core.TaskBase
        the Pydra object to convert to a dictionary
    required_modules : set[str]
        a set of modules that are required to load the pydra object back
        out from disk and run it
    workflow : pydra.Workflow, optional
        the containing workflow that the object to serialised is part of

    Returns
    -------
    dict
        the dictionary containing the contents of the Pydra object
    """
    dct = {
        "name": obj.name,
        "class": "<" + ClassResolver.tostr(obj, strip_prefix=False) + ">",
    }
    if isinstance(obj, Workflow):
        dct["nodes"] = [
            pydra_asdict(n, required_modules=required_modules, workflow=obj)
            for n in obj.nodes
        ]
        dct["outputs"] = outputs = {}
        for outpt_name, lf in obj._connections:
            outputs[outpt_name] = {"task": lf.name, "field": lf.field}
    else:
        if isinstance(obj, FunctionTask):
            func = cp.loads(obj.inputs._func)
            module = inspect.getmodule(func)
            dct["class"] = "<" + module.__name__ + ":" + func.__name__ + ">"
            required_modules.add(module.__name__)
            # inspect source for any import lines (should be present in function
            # not module)
            for line in inspect.getsourcelines(func)[0]:
                if match := extract_import_re.match(line):
                    required_modules.add(match.group(1))
            # TODO: check source for references to external modules that aren't
            #       imported within function
        elif type(obj).__module__ != "pydra.engine.task":
            pkg = package_from_module(type(obj).__module__)
            dct["package"] = pkg.key
            dct["version"] = pkg.version
        if hasattr(obj, "container"):
            dct["container"] = {"type": obj.container, "image": obj.image}
    dct["inputs"] = inputs = {}
    for inpt_name in obj.input_names:
        if not inpt_name.startswith("_"):
            inpt_value = getattr(obj.inputs, inpt_name)
            if isinstance(inpt_value, LazyField):
                inputs[inpt_name] = {"field": inpt_value.field}
                # If the lazy field comes from the workflow lazy in, we omit
                # the "task" item
                if workflow is None or inpt_value.name != workflow.name:
                    inputs[inpt_name]["task"] = inpt_value.name
            elif inpt_value == attrs.NOTHING:
                inputs[inpt_name] = NOTHING_STR
            else:
                inputs[inpt_name] = inpt_value
    return dct


def lazy_field_fromdict(dct: dict, workflow: Workflow):
    """Unserialises a LazyField object from a dictionary"""
    if "task" in dct:
        inpt_task = getattr(workflow, dct["task"])
        lf = getattr(inpt_task.lzout, dct["field"])
    else:
        lf = getattr(workflow.lzin, dct["field"])
    return lf


def pydra_fromdict(dct: dict, workflow: Workflow = None, **kwargs) -> TaskBase:
    """Recreates a Pydra Task/Workflow from a dictionary object created by
    `pydra_asdict`

    Parameters
    ----------
    dct : dict
        dictionary representations of the object to recreate
    name : str
        name to give the object
    workflow : pydra.Workflow, optional
        the containing workflow that the object to recreate is connected to
    **kwargs
        additional keyword arguments passed to the pydra Object init method

    Returns
    -------
    pydra.engine.core.TaskBase
        the recreated Pydra object
    """
    klass = ClassResolver()(dct["class"])
    # Resolve lazy-field references to workflow fields
    inputs = {}
    for inpt_name, inpt_val in dct["inputs"].items():
        if inpt_val == NOTHING_STR:
            continue
        # Check for 'field' key in a dictionary val and convert to a
        # LazyField object
        if isinstance(inpt_val, dict) and "field" in inpt_val:
            inpt_val = lazy_field_fromdict(inpt_val, workflow=workflow)
        inputs[inpt_name] = inpt_val
    kwargs.update((k, v) for k, v in inputs.items() if k not in kwargs)
    if klass is Workflow:
        obj = Workflow(name=dct["name"], input_spec=list(dct["inputs"]), **kwargs)
        for node_dict in dct["nodes"]:
            obj.add(pydra_fromdict(node_dict, workflow=obj))
        obj.set_output(
            [
                (n, lazy_field_fromdict(f, workflow=obj))
                for n, f in dct["outputs"].items()
            ]
        )
    else:
        obj = klass(name=dct["name"], **kwargs)
    return obj


@attrs.define
class ObjectConverter:

    klass: type
    allow_none: bool = False
    default_if_none: ty.Any = None
    accept_metadata: bool = False

    def __call__(self, value):
        return self._create_object(value)

    def _create_object(self, value, **kwargs):
        if value is None:
            if kwargs:
                value = {}
            elif self.allow_none:
                if callable(self.default_if_none):
                    default = self.default_if_none()
                else:
                    default = self.default_if_none
                return default
            else:
                raise ValueError(
                    f"None values not accepted in automatic conversion to {self.klass}"
                )
        if isinstance(value, dict):
            if self.accept_metadata:
                klass_attrs = set(attrs.fields_dict(self.klass))
                value_kwargs = {k: v for k, v in value.items() if k in klass_attrs}
                value_kwargs["metadata"] = {
                    k: v for k, v in value.items() if k not in klass_attrs
                }
            else:
                value_kwargs = value
            value_kwargs.update(kwargs)
            try:
                obj = self.klass(**value_kwargs)
            except TypeError as e:
                msg = f"when creating {self.klass} from {value_kwargs}"
                add_exc_note(e, msg)
                raise
        elif isinstance(value, (list, tuple)):
            obj = self.klass(*value, **kwargs)
        elif isinstance(value, self.klass):
            obj = copy(value)
            for k, v in kwargs.items():
                setattr(obj, k, v)
        elif isinstance(value, (str, int, float, bool)):
            # If there are kwargs that are in the first N positions of the
            # argument list, add them in as positional arguments first and then
            # append the value to the end of the args list
            args = []
            kgs = copy(kwargs)
            for field_name in attrs.fields_dict(self.klass):
                try:
                    args.append(kgs.pop(field_name))
                except KeyError:
                    break
            args.append(value)
            obj = self.klass(*args, **kgs)
        else:
            raise ValueError(f"Cannot convert {value} into {self.klass}")
        return obj


@attrs.define
class ObjectListConverter(ObjectConverter):
    def __call__(self, value):
        converted = []
        if isinstance(value, dict):
            for name, item in value.items():
                converted.append(self._create_object(item, name=name))
        else:
            for item in value:
                converted.append(self._create_object(item))
        return converted

    @classmethod
    def asdict(cls, objs: list, **kwargs) -> dict:
        dct = {}
        for obj in objs:
            obj_dict = attrs.asdict(obj, **kwargs)
            dct[obj_dict.pop("name")] = obj_dict
        return dct


def parse_value(value):
    """Parses values from string representations"""
    try:
        value = json.loads(
            value
        )  # FIXME: Is this value replace really necessary, need to investigate where it is used again
    except (TypeError, json.decoder.JSONDecodeError):
        pass
    return value



from abc import abstractmethod, ABCMeta
import re
import typing as ty
import attrs
from operator import attrgetter
from attrs.converters import optional

# from arcana.core.data.row import DataRow
from arcana.core.utils.serialize import ClassResolver
from arcana.core.exceptions import ArcanaDataMatchError
from ..analysis.salience import DataQuality, ColumnSalience
from .space import DataSpace


ItemType = ty.TypeVar("ItemType")


@attrs.define
class DataColumn(ty.Generic[ItemType], metaclass=ABCMeta):

    name: str = attrs.field()
    path: str = attrs.field()
    datatype = attrs.field()
    row_frequency: DataSpace = attrs.field()
    dataset = attrs.field(
        default=None, metadata={"asdict": False}, eq=False, hash=False, repr=False
    )

    def __iter__(self):
        return (n[self.name] for n in self.dataset.rows(self.row_frequency))

    def __getitem__(self, id) -> ItemType:
        return self.dataset.row(id=id, row_frequency=self.row_frequency)[self.name]

    def __len__(self):
        return len(list(self.dataset.rows(self.row_frequency)))

    @property
    def ids(self):
        return [n.id for n in self.dataset.rows(self.row_frequency)]

    @abstractmethod
    def match(self, row):
        """Selects a single item from a data row that matches the
        criteria/path of the column.

        Parameters
        ----------
        row: DataRow
            the row to match the item from

        Returns
        -------
        DataType
            the data item that matches the criteria/path

        Raises
        ------
        ArcanaDataMatchError
            if none or multiple items match the criteria/path of the column
            within the row
        ArcanaFileFormatError
            if there are no files matching the format of the column in the row"""

    def assume_exists(self):
        # Update local cache of sink paths
        for item in self:
            item.get(assume_exists=True)


@attrs.define
class DataSource(DataColumn):
    """
    Specifies the criteria by which an item is selected from a data row to
    be a data source.

    Parameters
    ----------
    path : str
        A regex name_path to match the file_group names with. Must match
        one and only one file_group per <row_frequency>. If None, the name
        is used instead.
    datatype : type
        File format that data will be
    row_frequency : DataSpace
        The row_frequency of the file-group within the dataset tree, e.g. per
        'session', 'subject', 'timepoint', 'group', 'dataset'
    quality_threshold : DataQuality
        The acceptable quality (or above) that should be considered. Data items
        will be considered missing
    order : int | None
        To be used to distinguish multiple file_groups that match the
        name_path in the same session. The order of the file_group within the
        session (0-indexed). Based on the scan ID but is more robust to small
        changes to the IDs within the session if for example there are
        two scans of the same type taken before and after a task.
    header_vals : Dict[str, str]
        To be used to distinguish multiple items that match the
        the other criteria. The provided dictionary contains
        header values that must match the stored header_vals exactly.
    is_regex : bool
        Flags whether the name_path is a regular expression or not
    """

    quality_threshold: DataQuality = attrs.field(
        default=None, converter=optional(lambda q: DataQuality[str(q)])
    )
    order: int = attrs.field(
        default=None, converter=lambda x: int(x) if x is not None else None
    )
    header_vals: ty.Dict[str, ty.Any] = attrs.field(default=None)
    is_regex: bool = attrs.field(
        default=False,
        converter=lambda x: x.lower() == "true" if isinstance(x, str) else x,
    )

    is_sink = False

    def match(self, row):
        criteria = [
            (match_path, self.path if not self.is_regex else None),
            (match_path_regex, self.path if self.is_regex else None),
            (match_quality, self.quality_threshold),
            (match_header_vals, self.header_vals),
        ]
        # Get all items that match the data type of the source
        matches = row.resolved(self.datatype)
        if not matches:
            msg = (
                f"Did not find any items matching data datatype "
                f"{ClassResolver.tostr(self.datatype)} in '{row.id}' "
                f"{self.row_frequency} for the "
                f"'{self.name}' column, found unresolved items:"
            )
            for item in sorted(row.unresolved, key=attrgetter("path")):
                msg += (
                    f"\n    {item.path}: paths="
                    + ",".join(p.name for p in item.file_paths)
                    + ((", uris=" + ",".join(item.uris.keys())) if item.uris else "")
                )
            msg += self._format_criteria()
            raise ArcanaDataMatchError(msg)
        # Apply all filters to find items that match criteria
        for func, arg in criteria:
            if arg is not None:
                filtered = [m for m in matches if func(m, arg)]
                if not filtered:
                    raise ArcanaDataMatchError(
                        "Did not find any items "
                        + func.__doc__.format(arg)
                        + self._error_msg(row, matches)
                    )
                matches = filtered
        # Select a single item from the ones that match the criteria
        if self.order is not None:
            try:
                match = matches[self.order - 1]
            except IndexError as e:
                raise ArcanaDataMatchError(
                    "Not enough matching items to select one at index "
                    f"{self.order} (starting from 1), found:"
                    + self._format_matches(matches)
                ) from e
        elif len(matches) > 1:
            raise ArcanaDataMatchError(
                "Found multiple matches " + self._error_msg(row, matches)
            )
        else:
            match = matches[0]
        return match

    def _error_msg(self, row, matches):
        return (
            f" attempting to select {ClassResolver.tostr(self.datatype)} item for "
            f"the '{row.id}' {row.frequency} in the '{self.name}' column, found:"
            + self._format_matches(matches)
            + self._format_criteria()
        )

    def _format_criteria(self):
        return (
            f"\n\n    criteria: {self.path}', is_regex={self.is_regex}, "
            f"datatype={ClassResolver.tostr(self.datatype)}, "
            f"quality_threshold='{self.quality_threshold}', "
            f"header_vals={self.header_vals}, order={self.order}"
        )

    def _format_matches(self, matches):
        out_str = ""
        for match in sorted(matches, key=attrgetter("path")):
            out_str += "\n    "
            if match.order:
                out_str += match.order + ": "
            out_str += match.path
            out_str += f" ({match.quality})"
        return out_str


def match_path(item, path):
    "at the path '{}'"
    return item.path == path


def match_path_regex(item, pattern):
    "that matched the path pattern '{}'"
    if not pattern.endswith("$"):
        pattern += "$"
    return re.match(pattern, item.path)


def match_quality(item, threshold):
    "with an acceptable quality '{}'"
    return item.quality >= threshold


def match_header_vals(item, header_vals):
    "with the header values '{}'"
    return all(item.header(k) == v for k, v in header_vals.items())


@attrs.define
class DataSink(DataColumn):
    """
    A specification for a file group within a analysis to be derived from a
    processing pipeline.

    Parameters
    ----------
    path : str
        The path to the relative location the corresponding data items will be
        stored within the rows of the data tree.
    datatype : type
        The file datatype or data type used to store the corresponding items
        in the store dataset.
    row_frequency : DataSpace
        The row_frequency of the file-group within the dataset tree, e.g. per
        'session', 'subject', 'timepoint', 'group', 'dataset'
    salience : Salience
        The salience of the specified file-group, i.e. whether it would be
        typically of interest for publication outputs or whether it is just
        a temporary file in a workflow, and stages in between
    pipeline_name : str
        The name of the workflow applied to the dataset to generates the data
        for the sink
    """

    salience: ColumnSalience = attrs.field(
        default=ColumnSalience.supplementary,
        converter=lambda s: ColumnSalience[str(s)] if s is not None else None,
    )
    pipeline_name: str = attrs.field(default=None)

    is_sink = True

    def match(self, row):
        matches = [i for i in row.resolved(self.datatype) if i.path == self.path]
        if not matches:
            # Return a placeholder data item th.datatypebe set
            return self.datatype(path=self.path, row=row, exists=False)
        elif len(matches) > 1:
            raise ArcanaDataMatchError(
                "Found multiple matches " + self._error_msg(row, matches)
            )
        return matches[0]

    def derive(self, ids=None):
        self.dataset.derive(self.name, ids=ids)



from __future__ import annotations
from types import MemberDescriptorType
from .spec import (
    _UnresolvedOp,
    _Inherited,
    _MappedFrom,
    ColumnSpec,
    Parameter,
    SubanalysisSpec,
)
from .make import make
from .salience import ColumnSalience, ParameterSalience, CheckSalience
from ..utils.misc import (
    SWICTH_ANNOTATIONS,
    CHECK_ANNOTATIONS,
    PIPELINE_ANNOTATIONS,
    CONVERTER_ANNOTATIONS,
)


def analysis(space: type):
    """Designate a class to be an analysis class

    Parameters
    ----------
    space : type (subclass of DataSpace)
        The data space the analysis operates on, see"""

    def decorator(klass):
        return make(klass, space)

    return decorator


def column(
    desc,
    row_frequency=None,
    salience=ColumnSalience.supplementary,
    metadata=None,
):
    return ColumnSpec(
        desc=desc,
        row_frequency=row_frequency,
        salience=salience,
        metadata=metadata,
    )


def parameter(
    desc,
    default=None,
    choices=None,
    lower_bound=None,
    upper_bound=None,
    salience=ParameterSalience.recommended,
    metadata=None,
):
    return Parameter(
        desc=desc,
        default=default,
        choices=choices,
        lower_bound=lower_bound,
        upper_bound=upper_bound,
        salience=salience,
        metadata=metadata,
    )


def subanalysis(desc, metadata=None, **mappings):
    return SubanalysisSpec(
        desc=desc, mappings=tuple(mappings.items()), metadata=metadata
    )


def pipeline(*outputs, condition=None, switch=None):
    """Decorate a instance method that adds nodes to an existing Pydra workflow

    Parameters
    ----------
    *outputs : list[column]
        outputs produced by the pipeline
    condition : Operation, optional
        condition on which the pipeline will be used instead of the default (the pipeline
        with condition is None)
    switch : str or tuple[str, Any], optional
        name of a "switch" method in the analysis class, which selects nodes to be run
        with this pipeline instead of the default. If a tuple, then the first element
        is the switch name and the second is the return value it should match (the
        return value should be boolean otherwise)
    """

    def decorator(meth):
        anots = meth.__annotations__[PIPELINE_ANNOTATIONS] = {}
        anots["outputs"] = outputs
        anots["condition"] = condition
        anots["switch"] = switch.__name__ if switch is not None else None
        return meth

    return decorator


def switch(meth):
    """Designates a method as being a "switch" that is used to determine which version
    of a pipeline is run

    Parameters
    ----------
    in_task : bool
        whether to wrap the switch in its own task or whether it adds its own nodes
        explicitly"""
    anot = meth.__annotations__
    anot[SWICTH_ANNOTATIONS] = True
    return meth


def check(column, salience=CheckSalience.prudent):
    """Decorate a method, which adds a quality control check to be run against a column"""

    def decorator(meth):
        meth.__annotations__[CHECK_ANNOTATIONS] = {
            "column": column,
            "salience": salience,
        }
        return meth

    return decorator


def inherit(ref: MemberDescriptorType = None, **to_overwrite):
    """Used to explicitly inherit a column or attribute from a base class so it can be
    used in a sub class. This explicit inheritance is enforced when the column/parameter
    is referenced in the base class in order to make the code more readable (i.e. so
    other developers can track where columns/parameters are defined)

    Parameters
    ----------
    ref : MemberDescriptorType, optional
        a reference to the field that is being inherited from. Note that it is not
        actually used for anything, the field to be inherited is determined by scanning
        the method-resolution order for matching names, but it can make the code more
        readable by linking the inherited attribute with its initial definition.
    **to_overwrite:
        any attributes to override from the inherited column/parameter
    """
    if "row_frequency" in to_overwrite:
        raise ValueError("Cannot overwrite row_frequency when inheriting")
    return _Inherited(to_overwrite)


def map_from(subanalysis_name, column_name, **to_overwrite):
    """Used to explicitly inherit a column or attribute from a base class so it can be
    used in a sub class. This explicit inheritance is enforced when the column/parameter
    is referenced in the base class in order to make the code more readable (i.e. so
    other developers can track where columns/parameters are defined)

    Parameters
    ----------
    base_class : type
        the base class to inherit the column/parameter from. The name will be matched
        to the name of the column/parameter in the base class
    **to_overwrite:
        any attributes to override from the inherited column/parameter
    """
    if "row_frequency" in to_overwrite:
        raise ValueError("Cannot overwrite row_frequency when mapping")
    return _MappedFrom(subanalysis_name, column_name, to_overwrite)


def value_of(param):
    """Specifies that the value of the parameter in question should be returned so it
    can be tested within a condition"""
    return _UnresolvedOp("value_of", (param,))


def is_provided(column, in_format: type = None):
    """Test whether a column mapping is specified when the analysis class is applied to a
    dataset"""
    return _UnresolvedOp(
        "is_provided",
        (
            column,
            in_format,
        ),
    )


def converter(output_format):
    def decorator(meth):
        anot = meth.__annotations__
        anot[CONVERTER_ANNOTATIONS] = output_format
        return meth

    return decorator



from __future__ import annotations
from pathlib import Path
import typing as ty
import attrs
from collections import defaultdict
from abc import ABCMeta
from arcana.core.exceptions import (
    ArcanaNameError,
    ArcanaWrongFrequencyError,
    ArcanaFileFormatError,
)
from .type.base import DataType
from ..analysis.salience import DataQuality
from .space import DataSpace

if ty.TYPE_CHECKING:
    import arcana.core.data.set


@attrs.define(auto_detect=True)
class DataRow:
    """A "row" in a dataset "frame" where file-groups and fields can be placed, e.g.
    a session or subject.

    Parameters
    ----------
    ids : Dict[DataSpace, str]
        The ids for the frequency of the row and all "parent" frequencies
        within the tree
    frequency : DataSpace
        The frequency of the row
    dataset : Dataset
        A reference to the root of the data tree
    """

    ids: ty.Dict[DataSpace, str] = attrs.field()
    frequency: DataSpace = attrs.field()
    dataset: arcana.core.data.set.Dataset = attrs.field(repr=False)
    children: ty.DefaultDict[
        DataSpace, ty.Dict[ty.Union[str, ty.Tuple[str]], str]
    ] = attrs.field(factory=lambda: defaultdict(dict), repr=False)
    _unresolved = attrs.field(default=None, repr=False)
    _items = attrs.field(factory=dict, init=False, repr=False)

    def __getitem__(self, column_name):
        """Gets the item for the current row

        Parameters
        ----------
        column_name : str
            Name of a selected column in the dataset

        Returns
        -------
        DataType
            The item matching the provided name specified by the column name
        """
        if column_name in self._items:
            return self._items[column_name]
        else:
            try:
                spec = self.dataset[column_name]
            except KeyError as e:
                raise ArcanaNameError(
                    column_name,
                    f"{column_name} is not the name of a column in "
                    f"{self.dataset.id} dataset ('"
                    + "', '".join(self.dataset.columns)
                    + "')",
                ) from e
            if spec.row_frequency != self.frequency:
                return ArcanaWrongFrequencyError(
                    column_name,
                    f"'column_name' ({column_name}) is of {spec.row_frequency} "
                    f"frequency and therefore not in rows of {self.frequency}"
                    " frequency",
                )
            item = self._items[column_name] = spec.match(self)
            return item

    def __setitem__(self, column_name, value):
        item = self[column_name]
        item.put(value)
        return item

    def __repr__(self):
        return f"{type(self).__name__}(id={self.id}, frequency={self.frequency})"

    @property
    def id(self):
        return self.ids[self.frequency]

    @property
    def label(self):
        return self.path[-1]

    def __iter__(self):
        return iter(self.keys())

    def keys(self):
        return (n for n, _ in self.items())

    def values(self):
        return (i for _, i in self.items())

    def items(self):
        return (
            (c.name, self[c.name])
            for c in self.dataset.columns.values()
            if c.row_frequency == self.frequency
        )

    def column_items(self, column_name):
        """Gets the item for the current row if item's frequency matches
        otherwise gets all the items that are related to the current row (
        i.e. are in child rows)

        Parameters
        ----------
        column_name : str
            Name of a selected column in the dataset

        Returns
        -------
        Sequence[DataType]
            The item matching the provided name specified by the column name
            if the column is of matching or ancestor frequency, or list of
            items if a descendent or unrelated frequency.
        """
        try:
            return [self[column_name]]
        except ArcanaWrongFrequencyError:
            # If frequency is not a ancestor row then return the
            # items in the children of the row (if they are child
            # rows) or the whole dataset
            spec = self.dataset.columns[column_name]
            try:
                return self.children[spec.row_frequency].values()
            except KeyError:
                return self.dataset.column(spec.row_frequency)

    @property
    def unresolved(self):
        if self._unresolved is None:
            self._unresolved = []
            self.dataset.store.find_items(self)
        return self._unresolved

    def resolved(self, datatype):
        """
        Items in the row that are able to be resolved to the given datatype

        Parameters
        ----------
        datatype : type
            The file datatype or type to reolve the item to
        """
        matches = []
        for potential in self.unresolved:
            try:
                matches.append(datatype.resolve(potential))
            except ArcanaFileFormatError:
                pass
        return matches

    @property
    def ids_tuple(self):
        return self.dataset.ids_tuple(self.ids)

    def add_file_group(self, path, **kwargs):
        if self._unresolved is None:
            self._unresolved = []
        self._unresolved.append(UnresolvedFileGroup(path=path, row=self, **kwargs))

    def add_field(self, path, value, **kwargs):
        if self._unresolved is None:
            self._unresolved = []
        self._unresolved.append(
            UnresolvedField(path=path, row=self, value=value, **kwargs)
        )


@attrs.define
class UnresolvedDataType(metaclass=ABCMeta):
    """A file-group stored in, potentially multiple, unknown file formats.
    File formats are resolved by providing a list of candidates to the
    'resolve' method

    Parameters
    ----------
    path : str
        The name_path to the relative location of the file group, i.e. excluding
        information about which row in the data tree it belongs to
    order : int | None
        The ID of the file_group in the session. To be used to
        distinguish multiple file_groups with the same scan type in the
        same session, e.g. scans taken before and after a task. For
        datasets where this isn't stored (i.e. Local), id can be None
    quality : DataQuality
        The quality label assigned to the file_group (e.g. as is saved on XNAT)
    provenance : Provenance | None
        The provenance for the pipeline that generated the file-group,
        if applicable
    """

    path: str = attrs.field(default=None)
    row: DataRow = attrs.field(default=None)
    order: int = attrs.field(default=None)
    quality: DataQuality = attrs.field(default=DataQuality.usable)
    provenance: ty.Dict[str, ty.Any] = attrs.field(default=None)
    _matched: ty.Dict[str, DataType] = attrs.field(factory=dict, init=False)

    @property
    def item_kwargs(self):
        return {
            "path": self.path,
            "order": self.order,
            "row": self.row,
            "quality": self.quality,
        }


def normalise_paths(file_paths):
    "Convert all file paths to absolute real paths"
    if file_paths:
        file_paths = [Path(p).absolute() for p in file_paths]
    return file_paths


@attrs.define
class UnresolvedFileGroup(UnresolvedDataType):
    """A file-group stored in, potentially multiple, unknown file formats.
    File formats are resolved by providing a list of candidates to the
    'resolve' method

    Parameters
    ----------
    name_path : str
        The name_path to the relative location of the file group, i.e. excluding
        information about which row in the data tree it belongs to
    order : int | None
        The ID of the file_group in the session. To be used to
        distinguish multiple file_groups with the same scan type in the
        same session, e.g. scans taken before and after a task. For
        datasets where this isn't stored (i.e. Local), id can be None
    quality : DataQuality
        The quality label assigned to the file_group (e.g. as is saved on XNAT)
    provenance : Provenance | None
        The provenance for the pipeline that generated the file-group,
        if applicable
    row : DataRow
        The data row that the field belongs to
    file_paths : Sequence[str] | None
        Path to the file-group in the local cache
    uris : Dict[str, str] | None
        For stores where the name of the file datatype is saved with the
        data (i.e. XNAT), the name of the resource enables straightforward
        datatype identification. It is stored here along with URIs corresponding
        to each resource
    """

    file_paths: ty.Sequence[Path] = attrs.field(factory=list, converter=normalise_paths)
    uris: ty.Dict[str] = attrs.field(default=None)

    @classmethod
    def from_paths(cls, base_dir: Path, paths: ty.List[Path], **kwargs):
        groups = defaultdict(list)
        for path in paths:
            relpath = path.relative_to(base_dir)
            path_stem = str(relpath)[: -len("".join(relpath.suffixes))]
            groups[path_stem].append(path)  # No extension case
            # Add all possible stems
            for i in range(len(relpath.suffixes)):
                groups["".join([path_stem] + relpath.suffixes[: (i + 1)])].append(path)
        return [cls(path=p, file_paths=g, **kwargs) for p, g in groups.items()]


@attrs.define
class UnresolvedField(UnresolvedDataType):
    """A file-group stored in, potentially multiple, unknown file formats.
    File formats are resolved by providing a list of candidates to the
    'resolve' method

    Parameters
    ----------
    path : str
        The name_path to the relative location of the file group, i.e. excluding
        information about which row in the data tree it belongs to
    value : str
        The value assigned to the unresolved data item (for fields instead of
        file groups)
    order : int | None
        The ID of the file_group in the session. To be used to
        distinguish multiple file_groups with the same scan type in the
        same session, e.g. scans taken before and after a task. For
        datasets where this isn't stored (i.e. Local), id can be None
    quality : DataQuality
        The quality label assigned to the file_group (e.g. as is saved on XNAT)
    provenance : Provenance | None
        The provenance for the pipeline that generated the file-group,
        if applicable
    row : DataRow
        The data row that the field belongs to
    """

    value: ty.Union[
        float, int, str, ty.List[float], ty.List[int], ty.List[str]
    ] = attrs.field(default=None)

    # def _resolve(self, datatype):
    #     try:
    #         if datatype._name == 'Sequence':
    #             if len(datatype.__args__) > 1:
    #                 raise ArcanaUsageError(
    #                     f"Sequence formats with more than one arg "
    #                     "are not supported ({datatype})")
    #             subtype = datatype.__args__[0]
    #             value = [subtype(v)
    #                         for v in self.value[1:-1].split(',')]
    #         else:
    #             value = datatype(self.value)
    #     except ValueError as e:
    #         raise ArcanaUnresolvableFormatException(
    #                 f"Could not convert value of {self} ({self.value}) "
    #                 f"to datatype {datatype}") from e
    #     else:
    #         item = DataType(value=value, **self.item_kwargs)
    #     return item



import os.path
import traceback


def show_cli_trace(result):
    return "".join(traceback.format_exception(*result.exc_info))


def make_dataset_locator(dataset, name=None):
    id_str = "file//" + os.path.abspath(dataset.id)
    if name is not None:
        id_str += "@" + name
    return id_str
